"""
toil_mergevcfs commands tests.

tmpdir is a py.path.local, learn: https://py.readthedocs.io/en/latest/path.html
"""
import gzip
from os.path import join

from toil_mergevcfs import commands
from toil_mergevcfs import __version__
from tests.utils import TEST


EXPECTED_SOMATIC_SNVS_1 = "8\t332591\t.\tC\tG"
EXPECTED_SOMATIC_SNVS_2 = "8\t9184628\t8421b6e2-505f-11e7-bd98-d60f982b8ef5\tA\tG"
EXPECTED_SOMATIC_SNVS_3 = "8\t17945818\t8421bc64-505f-11e7-bd98-d60f982b8ef5\tA\tG"
EXPECTED_SOMATIC_SNVS_4 = "8\t21417475\t8421bdfe-505f-11e7-bd98-d60f982b8ef5\tC\tT"

EXPECTED_SOMATIC_INDEL_1 = "8\t4291361\ta4f81b06-505d-11e7-be92-c33fa51be26e\tG\tGA"
EXPECTED_SOMATIC_INDEL_2 = "8\t6710087\t9efc0474-505d-11e7-be92-c33fa51be26e\tTC\tT"

EXPECTED_GERMLINE_SNVS_1 = "2\t796\t.\tT\tA"
EXPECTED_GERMLINE_SNVS_2 = "2\t123881\t.\tT\tC"
EXPECTED_GERMLINE_SNVS_3 = "2\t121\t.\tC\tT"
EXPECTED_GERMLINE_SNVS_4 = "2\t111927\t.\tA\tG"

EXPECTED_GERMLINE_INDEL_1 = "2\t123513\t.\tT\tTG"
EXPECTED_GERMLINE_INDEL_2 = "2\t123601\t.\tTG\tT"


def get_variant_lines(content):
    """Count the number of variants."""
    return len([v for v in content.strip().split("\n") if not v.startswith("#")])


def test_somatic_indels(tmpdir):
    """Test of somatic_indels vcf merge."""
    vcf_list = [TEST["pindel_indels"], TEST["strelka_indels"], TEST["mutect_indels"]]
    expected_variants = [EXPECTED_SOMATIC_INDEL_1, EXPECTED_SOMATIC_INDEL_2]
    run_main_command(
        tmpdir=tmpdir,
        vcf_list=vcf_list,
        expected_variants=expected_variants,
        not_expected_variants=[],
    )


def test_somatic_snvs(tmpdir):
    """Test of somatic_snvs vcf merge."""
    vcf_list = [TEST["caveman_snvs"], TEST["strelka_snvs"], TEST["mutect_snvs"]]
    expected_variants = [
        EXPECTED_SOMATIC_SNVS_1,
        EXPECTED_SOMATIC_SNVS_2,
        EXPECTED_SOMATIC_SNVS_3,
        EXPECTED_SOMATIC_SNVS_4,
    ]
    run_main_command(
        tmpdir=tmpdir,
        vcf_list=vcf_list,
        expected_variants=expected_variants,
        not_expected_variants=[],
    )


def test_somatic_regions(tmpdir):
    """Test of somatic_snvs vcf merge."""
    vcf_list = [TEST["caveman_snvs"], TEST["strelka_snvs"], TEST["mutect_snvs"]]
    regions_file = TEST["regions_file_8"]
    expected_variants = [EXPECTED_SOMATIC_SNVS_2]
    not_expected_variants = [
        EXPECTED_SOMATIC_SNVS_1,
        EXPECTED_SOMATIC_SNVS_3,
        EXPECTED_SOMATIC_SNVS_4,
    ]
    run_main_command(
        tmpdir=tmpdir,
        vcf_list=vcf_list,
        expected_variants=expected_variants,
        not_expected_variants=not_expected_variants,
        regions_file=regions_file,
    )


def test_germline_indels(tmpdir):
    """Test of germline_indels vcf merge."""
    vcf_list = [TEST["freebayes_indels"], TEST["strelka_gm_indels"]]
    expected_variants = [EXPECTED_GERMLINE_INDEL_1, EXPECTED_GERMLINE_INDEL_2]
    run_main_command(
        tmpdir=tmpdir,
        vcf_list=vcf_list,
        expected_variants=expected_variants,
        not_expected_variants=[],
    )


def test_germline_snvs(tmpdir):
    """Test of germline_snvs vcf merge."""
    vcf_list = [TEST["freebayes_snvs"], TEST["strelka_gm_snvs"]]
    expected_variants = [
        EXPECTED_GERMLINE_SNVS_1,
        EXPECTED_GERMLINE_SNVS_2,
        EXPECTED_GERMLINE_SNVS_3,
        EXPECTED_GERMLINE_SNVS_4,
    ]
    run_main_command(
        tmpdir=tmpdir,
        vcf_list=vcf_list,
        expected_variants=expected_variants,
        not_expected_variants=[],
    )


def test_germline_regions(tmpdir):
    """Test of germline_snvs vcf merge."""
    vcf_list = [TEST["freebayes_snvs"], TEST["strelka_gm_snvs"]]
    regions_file = TEST["regions_file_2"]
    expected_variants = [EXPECTED_GERMLINE_SNVS_1, EXPECTED_GERMLINE_SNVS_3]
    not_expected_variants = [EXPECTED_GERMLINE_SNVS_2, EXPECTED_GERMLINE_SNVS_4]
    run_main_command(
        tmpdir=tmpdir,
        vcf_list=vcf_list,
        expected_variants=expected_variants,
        not_expected_variants=not_expected_variants,
        regions_file=regions_file,
    )


def run_main_command(
    tmpdir,
    vcf_list=None,
    expected_variants=None,
    not_expected_variants=None,
    regions_file=None,
):
    """Sample test for the main command."""
    outdir = tmpdir.strpath
    jobstore = join(outdir, "jobstore")
    outfile = join(outdir, "merged.vcf.gz")

    args = [
        jobstore,
        "--stats",
        "--disableCaching",
        "--realTimeLogging",
        "--writeLogs",
        join(outdir, "logs_toil"),
        "--logFile",
        join(outdir, "head_job.toil"),
        "--logLevel",
        "ERROR",
        "--out",
        outfile,
    ]
    if regions_file:
        args += ["--regions-file", regions_file]
    for vcf in vcf_list:
        args += ["--vcf", vcf]

    # Get and validate options.
    parser = commands.get_parser()
    options = parser.parse_args(args)
    options = commands.process_parsed_options(options)

    # Call pipeline
    commands.run_toil(options)

    # Check if the output files are correct
    initial_variant_count = sum([get_variant_lines(vcf) for vcf in vcf_list])
    final_variants_count = get_variant_lines(outfile)
    assert final_variants_count > 0
    assert final_variants_count <= initial_variant_count

    # Check if the output files are correct
    with gzip.open(outfile, "rb") as f:
        content = f.read()
        for expected_variant in expected_variants or []:
            assert expected_variant in content, outfile
        for not_expected_variant in not_expected_variants or []:
            assert not_expected_variant not in content, outfile
