"""toil_mergevcfs commands."""
from glob import glob
from os.path import basename, isdir, join
import tempfile
import subprocess

import click
from toil_container import ContainerArgumentParser
from toil_container import ContainerJob
from pysam import VariantFile

from toil_mergevcfs import __version__
from toil_mergevcfs.utils import add_passed_field, add_version, rename_samples_headers


class PreprocessVcf(ContainerJob):

    """Job class to check vcfs are normalized, gzipped and tabixed."""

    def subset_vcf(self, fileStore, in_vcf, out_vcf):
        """Filter out regions that are not in the bedfile."""
        out_vcf = out_vcf.strip(".gz")
        cmd = [
            "bcftools",
            "view",
            "--regions-file",
            self.options.regions_file,
            "--output-file",
            out_vcf,
            in_vcf,
        ]

        fileStore.logToMaster(" ".join(cmd))
        self.call(["bash", "-c", " ".join(cmd)])

        cmd = ["bgzip", "-f", out_vcf]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)
        out_vcf += ".gz"

        cmd = ["tabix", "-fp", "vcf", out_vcf]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)

    def normalize_vcf(self, fileStore, in_vcf, out_vcf):
        """Normalize and realign indels and split multi-allelic rows."""
        out_vcf = out_vcf.strip(".gz")

        cmd = ["bcftools", "norm", "-m", "-"]
        if self.options.reference:
            cmd += ["--fasta-ref", self.options.reference]
        cmd += ["--output", out_vcf, in_vcf]

        fileStore.logToMaster(" ".join(cmd))
        self.call(["bash", "-c", " ".join(cmd)])

        cmd = ["bgzip", "-f", out_vcf]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)
        out_vcf += ".gz"

        cmd = ["tabix", "-fp", "vcf", out_vcf]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)

    def run(self, fileStore):
        """Preprocess vcfs before merge."""
        for vcf in self.options.vcf:
            vcf_name = basename(vcf)
            in_vcf = vcf

            if self.options.regions_file:
                # filter vcf with regions contained in the file
                out_vcf = join(self.options.working_dir, "subset_{}".format(vcf_name))
                self.subset_vcf(fileStore, in_vcf=in_vcf, out_vcf=out_vcf)
                in_vcf = out_vcf

            # add 'PASSED' field under INFO. ex: PASSED_caveman,PASSED_mutect
            out_vcf = join(self.options.working_dir, "passed_{}".format(vcf_name))
            add_passed_field(in_vcf=in_vcf, out_vcf=out_vcf)
            in_vcf = out_vcf

            # modify column sample name to include the caller name. ex: strelka_NORMAL.
            out_vcf = join(self.options.working_dir, "renamed_{}".format(vcf_name))
            rename_samples_headers(in_vcf=in_vcf, out_vcf=out_vcf)
            in_vcf = out_vcf

            # Normalize by realign indels and split multi-allelic rows.
            out_vcf = join(self.options.working_dir, "norm_{}".format(vcf_name))
            self.normalize_vcf(fileStore, in_vcf=in_vcf, out_vcf=out_vcf)
            in_vcf = out_vcf


class MergeByRegion(ContainerJob):

    """A job class to merge vcfs by regions."""

    def __init__(self, options=None, *args, **kwargs):
        """Set variables that will be used in later steps."""
        self.options = options
        self.region = self.options.region
        super(MergeByRegion, self).__init__(options=self.options, *args, **kwargs)

    def run(self, fileStore):
        """Split vcf by contigs in the vcf."""
        outfile = basename(self.options.out)
        output_file = join(
            self.options.working_dir,
            "regions",
            outfile.strip(".gz").replace(".vcf", ".{}.vcf".format(self.region)),
        )
        input_files = [
            join(self.options.working_dir, "norm_{}".format(basename(vcf)))
            for vcf in self.options.vcf
        ]
        cmd = [
            "bcftools",
            "merge",
            "-m",
            "none",
            "--threads",
            str(self.options.threads),
            "--regions",
            self.region,
            "--output",
            output_file,
        ] + input_files

        fileStore.logToMaster(" ".join(cmd))
        self.call(["bash", "-c", " ".join(cmd)])

        cmd = ["bgzip", "-f", output_file]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)
        output_file += ".gz"

        cmd = ["tabix", "-fp", "vcf", output_file]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)


class ConcatenateRegions(ContainerJob):

    """A job class to concatenate the splitted vcfs."""

    def run(self, fileStore):
        """Run concat for vcf files."""
        region_files = glob(join(self.options.working_dir, "regions", "*.vcf.gz"))
        output_file = self.options.out.strip(".gz")
        cmd = [
            "bcftools",
            "concat",
            "--remove-duplicates",
            "--allow-overlaps",
            "--output",
            output_file,
        ] + region_files

        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)

        cmd = ["bgzip", "-f", output_file]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)
        output_file += ".gz"

        cmd = ["tabix", "-fp", "vcf", output_file]
        fileStore.logToMaster(" ".join(cmd))
        self.call(cmd)

        add_version(output_file)


def get_parser():
    """Get pipeline configuration using toil's."""
    parser = ContainerArgumentParser(
        version=__version__,
        description="A toil tool to merge vcfs from different callers.",
    )

    settings = parser.add_argument_group("pipeline arguments")

    settings.add_argument(
        "--vcf",
        help="Path of target bam file.",
        required=True,
        action="append",
        type=click.Path(file_okay=True, readable=True, resolve_path=True),
    )
    settings.add_argument(
        "--out",
        help="Output directory",
        required=True,
        type=click.Path(dir_okay=True, writable=True, resolve_path=True),
    )
    settings.add_argument(
        "--reference",
        help="Reference .fasta used to left-align indels of vcfs.",
        required=False,
        type=click.Path(file_okay=True, readable=True, resolve_path=True),
    )
    settings.add_argument(
        "--regions-file",
        help="Bed file to subset the files before merging",
        required=False,
        type=click.Path(file_okay=True, readable=True, resolve_path=True),
    )
    settings.add_argument(
        "--threads", default=4, help="Number of extra output compression threads."
    )

    return parser


def process_parsed_options(options):
    """Perform validations and add post parsing attributes to `options`."""
    if not len(options.vcf) > 1:
        raise click.UsageError("You have to pass more than one --vcf to merge.")

    if options.writeLogs is not None:
        subprocess.check_call(["mkdir", "-p", options.writeLogs])

    options.working_dir = temp = tempfile.mkdtemp()
    if not isdir(temp):
        subprocess.check_call(["mkdir", "-p", temp])

    regions_outdir = join(temp, "regions")
    if not isdir(regions_outdir):
        subprocess.check_call(["mkdir", "-p", regions_outdir])
    return options


def run_toil(options):
    """Run toil pipeline give an options namespace."""
    head_job = PreprocessVcf(cores=1, memory="1G", options=options)
    # head_job = ContainerJob(cores=1, memory="1G", options=options)
    concat_job = ConcatenateRegions(options=options, cores=1, memory="1G", runtime=60)
    contigs = set(
        [contig for vcf in options.vcf for contig in VariantFile(vcf).header.contigs]
    )
    for region in contigs:
        options.region = region
        region_job = MergeByRegion(options=options, cores=1, memory="1G", runtime=60)
        head_job.addChild(region_job)
    head_job.addFollowOn(concat_job)

    ContainerJob.Runner.startToil(head_job, options)


def main():
    """
    Parse options and run toil.

    **Workflow**

    1. Define Options using `get_parser`: build an `arg_parse` object that
       includes both toil options and pipeline specific options. These will be
       separated in different sections of the `--help` text and used by the
       jobs to do the work.

    2. Validate with `process_parsed_options`: once the options are parsed, it
       maybe necessary to conduct *post-parsing* operations such as adding new
       attributes to the `options` namespace or validating combined arguments.

    3. Execute with `run_toil`: this function uses the `options` namespace to
       build and run the toil `DAG`.
    """
    options = get_parser().parse_args()
    options = process_parsed_options(options=options)
    run_toil(options=options)
