"""toil_mergevcfs utils."""
from os.path import basename
import os
import binascii
import gzip
import tempfile
import shutil
import subprocess

import click
from pysam import VariantFile
from toil_mergevcfs import __version__


def is_gz_file(f):
    """Return true if a given file is gziped."""
    with open(f, "rb") as fin:
        return binascii.hexlify(fin.read(2)) == b"1f8b"


def get_file_headers(filename):
    """Get the metadata headers of the vcf or tsv files."""
    open_fn = gzip.open if is_gz_file(filename) else open
    headers = []
    with open_fn(filename, "rb") as file_handler:
        for line in file_handler.readlines():
            if line.startswith(b"#"):
                headers.append(line.decode())
            else:
                break
    return headers


def get_caller(vcf):
    """Determine which caller produced a given vcf file."""
    content = "\n".join(get_file_headers(vcf)).lower()
    callers = ["mutect", "strelka", "caveman", "pindel", "freebayes"]
    caller = {caller for caller in callers if caller in content}

    if len(caller) != 1:
        sample_name = basename(vcf).split(".vcf")[0]
        msg = (
            "Unable to determine caller of {}: None or 1+ callers found. "
            "Using sample name {}."
        ).format(vcf, sample_name)
        click.echo(msg)
        return sample_name

    click.echo("{} is produced by {}".format(vcf, list(caller)[0]))
    return list(caller)[0]


def add_passed_field(in_vcf, out_vcf):
    """
    Add PASSED_{caller} fields.

    Add flags (e.g. PASSED_caveman) under INFO for PASS variant in aim of reduce
    ambiguity of confident variants in the merged vcf.
    """
    caller = get_caller(in_vcf)

    i_vcf = VariantFile(in_vcf, "rb")
    new_header = i_vcf.header.copy()
    try:
        new_header.info.add(
            "PASSED_{}".format(caller),
            ".",
            "Flag",
            "this variants passed which caller(s)",
        )
        i_vcf.header.info.add(
            "PASSED_{}".format(caller),
            ".",
            "Flag",
            "this variants passed which caller(s)",
        )
    except ValueError:
        pass

    raw_out = out_vcf.strip(".gz")
    o_vcf = VariantFile(raw_out, "w", header=new_header)

    for record in i_vcf:
        new_rec = record.copy()
        filters = list(record.filter)
        if filters and filters[0] == "PASS":
            new_rec.info["PASSED_{}".format(caller)] = 1
        o_vcf.write(new_rec)

    o_vcf.close()

    subprocess.check_call(["bgzip", "-f", raw_out])


def rename_samples_headers(in_vcf, out_vcf):
    """Replace hard-to-read and ambiguious header with clear header."""
    out_vcf = out_vcf.strip(".gz")
    caller = get_caller(in_vcf)
    vcf = VariantFile(in_vcf, "rb")
    samples_names = list(vcf.header.samples)
    header = "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT"

    if len(samples_names) == 2:
        header += "\t{}_NORMAL\t{}_TUMOR\n".format(caller, caller)
    elif len(samples_names) == 1:
        header += "\t{}_NORMAL\n".format(caller)

    with open(out_vcf, "w") as fout:
        with gzip.open(in_vcf, "r") as fin:
            for line in fin:
                if line.startswith("#CHROM"):
                    fout.write(header)
                else:
                    fout.write(line)

    subprocess.check_call(["bgzip", "-f", out_vcf])
    os.remove(in_vcf)


def add_version(in_vcf):
    """Add click_mergevcfs version in the output vcf header."""
    temp = tempfile.NamedTemporaryFile(suffix=".tmp.vcf", delete=False)
    if in_vcf.endswith("gz"):
        with gzip.open(in_vcf, "rb") as fin:
            lines = fin.readlines()
    else:
        with open(in_vcf, "r") as fin:
            lines = fin.readlines()
    lines.insert(1, "##toil_mergevcfs={}\n".format(__version__))
    with open(temp.name, "w") as fout:
        for line in lines:
            fout.write(line.decode("utf-8") if isinstance(line, bytes) else line)
    subprocess.check_call(["bgzip", temp.name])
    shutil.move(temp.name + ".gz", in_vcf)
